'''
 # @ Author: Andrew Hossack
 # @ Create Time: 2022-04-01 12:47:58
 # Thanks to https://mike.depalatis.net/blog/simplifying-argparse.html
'''
import argparse
import os
import sys as _sys
import webbrowser

from dashtools.cli import update
from dashtools.deploy import deployHeroku
from dashtools.docker import dockerUtils
from dashtools.runtime import runtimeUtils
from dashtools.templating import buildApp, buildAppUtils, createTemplate
from dashtools.version import __version__


class MyArgumentParser(argparse.ArgumentParser):
    """Override default help message"""

    def print_help(self, file=None):
        if file is None:
            file = _sys.stdout
        message = f"""The dashtools v{__version__} CLI for Plotly Dash. See https://github.com/andrew-hossack/dash-tools for more details.
\nUsage:
    {'dashtools <command> [options]':<29}
\nCommands and Options:
    {'docker':<29}Handle Docker creation. Choose option:
        {'--init <image name>':<25}Creates a docker image in current directory

    {'heroku':<29}Handle Heroku deployment. Choose option:
        {'--deploy':<25}Deploys the current project to Heroku
        {'--update [remote name]':<25}Push changes to existing Heroku remote

    {'init <app name> [template]':<29}Create a new app
        {'--dir, -d':<25}Specify alternative create location

    {'run':<29}Run the app (experimental)
        {'--set-py-cmd <command>':<25}Set the python shell command

    {'templates':<29}List and create templates
        {'--init <directory>':<25}Creates a template from specified directory
        {'--list':<25}List available templates
\nOther Options:
    {'--help, -h':<29}Display help message
    {'--report-issue':<29}Report a bug or issue
    {'--version, -v':<29}Display version
"""
        file.write(message+"\n")


parser = MyArgumentParser()

parser._positionals.title = 'Positional Arguments'
parser._optionals.title = 'Optional Arguments'

subparsers = parser.add_subparsers(dest="subcommand")


def argument(*name_or_flags, **kwargs):
    """Convenience function to properly format arguments to pass to the
    subcommand decorator.
    """
    return (list(name_or_flags), kwargs)


def subcommand(args=[], parent=subparsers):
    """Decorator to define a new subcommand in a sanity-preserving way.
    The function will be stored in the ``func`` variable when the parser
    parses arguments so that it can be called directly like so::
        args = cli.parse_args()
        args.func(args)
    Usage example::
        @subcommand([argument("-d", help="Enable debug mode", action="store_true")])
        def subcommand(args):
            print(args)
    Then on the command line::
        $ python cli.py subcommand -d
    """
    def decorator(func):
        parser = parent.add_parser(func.__name__, description=func.__doc__)
        for arg in args:
            parser.add_argument(*arg[0], **arg[1])
        parser.set_defaults(func=func)
    return decorator


parser.add_argument(
    '-v',
    '--version',
    action='version',
    version=f'dashtools {__version__}')

parser.add_argument(
    '--report-issue',
    action='store_true',
    help='Report a bug or issue')


@subcommand(
    [
        argument(
            '--init',
            help='Create a new docker image for the current project',
            metavar='<image name>',
            nargs=1
        ),
    ])
def docker(args):
    """Initialize a new docker image for the current project."""
    if args.init:
        dockerUtils.create_image(
            image_name=args.init[0],
            cwd=os.getcwd()
        )
    else:
        print('dashtools: docker: error: Too few arguments')
        exit('dashtools: Available docker options: --init [--dir, -d]')


@ subcommand(
    [
        argument(
            "init",
            help='Create a new Dash app. Args: REQUIRED: <app name> OPTIONAL: <template> (Default: "default").',
            metavar='<app name> [<template>]',
            nargs='+'),
        argument(
            '--dir',
            '-d',
            help='Specify the directory to create the app in. Args: REQUIRED: <directory>',
            default=os.getcwd(),
            metavar='<directory>',
            nargs='?')
    ])
def init(args):
    """Initialize a new app."""
    if args.dir is None:
        print('dashtools: init: No directory specified. Usage Example: dashtools init MyDashApp -d ~/Desktop')
        exit('dashtools: init: Failed')
    buildApp.create_app(
        target_dir=args.dir,
        app_name=args.init[0],
        template=buildAppUtils.get_template_from_args(args))
    print(
        f'dashtools: Run your app using the "python {os.path.join(args.init[0], "src", "app.py")}" command')
    print(f'dashtools: For an in-depth guide on configuring your app, see https://dash.plotly.com/layout')


@ subcommand(
    [
        argument(
            "--list",
            help="List available templates",
            default=False,
            action="store_true"),
        argument(
            "--init",
            help="Convert directory to a template. Args: REQUIRED: <directory to convert>",
            nargs=1),
    ])
def templates(args):
    if args.list:
        print('dashtools: Templates usage example, type: dashtools init MyApp csv')
        print('dashtools: templates: List of available templates:')
        buildAppUtils.print_templates()
    elif args.init:
        createTemplate.create_template(src=args.init[0], dest=os.getcwd())
    else:
        print('dashtools: templates error: too few arguments')
        print('dashtools: For more information on templates, see https://github.com/andrew-hossack/dash-tools#templates')
        exit('dashtools: Available templates options: --list, --init')


@ subcommand(
    [
        argument(
            "--deploy",
            help="Deploys the current project to Heroku. Run command from the root of the project",
            default=False,
            action="store_true",
        ),
        argument(
            "--update",
            help="Updates the current project to Heroku. Run command from the root of the project. Args: OPTIONAL: <git remote>",
            nargs="?",
            const="heroku",
            type=str
        ),
    ])
def heroku(args):
    if args.deploy:
        deployHeroku.deploy_app_to_heroku(os.getcwd())
    elif args.update:
        deployHeroku.update_heroku_app(os.getcwd(), remote=args.update)
    else:
        print('dashtools: heroku error: too few arguments')
        exit('dashtools: Available heroku options: --deploy, --update')


@ subcommand(
    [
        argument(
            "run",
            help="Run the app locally. Uses Procfile if available, else recursive search for app.py",
            default=False,
            action="store_true"),
        argument(
            '--set-py-cmd',
            help='Set the python shell command. Args: REQUIRED: <python shell command>',
            metavar='<python shell command>',
            nargs=1)
    ])
def run(args):
    if args.set_py_cmd:
        runtimeUtils.set_python_shell_cmd(args.set_py_cmd[0])
    elif args.run:
        try:
            runtimeUtils.run_app(os.getcwd())
        except RuntimeError as e:
            print(e)
            exit('dashtools: run: Failed')


def main():
    """
    dashtools CLI entry point.
    """
    args = parser.parse_args()
    if args.subcommand:
        args.func(args)
    elif args.report_issue:
        print('dashtools: Report an issue at: https://github.com/andrew-hossack/dash-tools/issues/new/choose')
        if input('dashtools: Open in browser? (y/n) > ') == 'y':
            webbrowser.open(
                'https://github.com/andrew-hossack/dash-tools/issues/new/choose')
    else:
        parser.print_help()
    update.check_for_updates()



'''
 # @ Author: Andrew Hossack
 # @ Create Time: 2022-05-22 22:21:30
 # @ Used to create new templates from a directory
'''

import os
import shutil

from dashtools.deploy import deployHeroku, fileUtils
from dashtools.templating import buildAppUtils


def create_template(src: os.PathLike, dest: os.PathLike):
    """
    Recursively copies all files and directories from src. Appends '.template' to all files.
    Saves new template to parent directory of src.
    """

    if not os.path.exists(src):
        print(f"dashtools: templates: init: Source file {src} does not exist")
        exit('dashtools: templates: init: Failed')

    # Check for file write permissions in the base directory (command invoke directory)
    if not buildAppUtils.check_write_permission(dest):
        print(
            f'dashtools: templates: init: No write permissions for {dest}')
        exit(f'dashtools: templates: init: Failed')

    # Create the new template directory
    src = os.path.join(src, '')
    template_base_dir = os.path.join(dest, f'{os.path.normpath(src)}Template')
    if os.path.exists(template_base_dir):
        print(
            f'dashtools: templates: init: Template directory {template_base_dir} already exists')
        exit(f'dashtools: templates: init: Failed')

    # Check Procfile exists
    if not fileUtils.check_file_exists(src, 'Procfile'):
        print(
            f'dashtools: templates: init: No Procfile found in {src}')
        if deployHeroku.prompt_user_choice("dashtools: templates: init: Create Procfile?"):
            fileUtils.create_procfile(src)

    # Verify procfile
    if not fileUtils.verify_procfile(src)["valid"]:
        print(
            f'dashtools: templates: init: Procfile in {src} is invalid')
        if not deployHeroku.prompt_user_choice("dashtools: Continue?"):
            exit(f'dashtools: templates: init: Aborted')

    # Check runtime exists
    if not fileUtils.check_file_exists(src, 'runtime.txt'):
        print(
            f'dashtools: templates: init: No runtime.py found in {src}')
        if deployHeroku.prompt_user_choice("dashtools: templates: init: Create runtime.py?"):
            fileUtils.create_runtime_txt(src)

    # Make the new template directory
    os.mkdir(template_base_dir)
    # Copy files from src to template_base_dir, appending '.template' to all files
    for path, _, files in os.walk(src):
        for name in files:
            # Skip already .template files
            if('.template' in name):
                continue

            # Get the relative path to the file
            relative_path = os.path.relpath(
                path, src)
            rel_src = os.path.join(path, name)

            # Get the destination path
            rel_dest = relative_path if relative_path != '.' else ''
            dest = os.path.join(
                template_base_dir,
                rel_dest,
                name +
                '.template')

            # Create the directory if it doesn't exist
            if not os.path.exists(os.path.dirname(dest)):
                os.makedirs(os.path.dirname(dest))

            # Copy the file
            shutil.copyfile(rel_src, dest)
    print(
        f'dashtools: templates: init: Created template {template_base_dir}')



'''
 # @ Author: Andrew Hossack
 # @ Create Time: 2022-04-28 11:43:00
 # Git util functions
'''
import os
import re
import subprocess


def git_is_installed() -> bool:
    """
    Check that git is installed on the system
    """
    regex = r'git version'
    isInstalled = True
    try:
        git_command_output = subprocess.check_output(
            'git --version', shell=True)
        if not re.search(regex, git_command_output.decode('utf-8')):
            isInstalled = False
    except subprocess.CalledProcessError:
        isInstalled = False
    return isInstalled


def is_git_repository() -> bool:
    """
    Check that the current location is a git repository or not
    by checking .git directory
    """
    isGitRepo = False
    if os.path.isdir('.git'):
        isGitRepo = True
    return isGitRepo



'''
 # @ Author: Andrew Hossack
 # @ Create Time: 2022-08-09 10:17:39
 # Checks for newer versions of dash-tools with yolk3k and prompt updates if available
'''


import re
import subprocess
from typing import Union

from dashtools.version import __version__
from termcolor import cprint


def _new_version_available() -> Union[str, None]:
    """
    Checks for newer version using yolk3k

    Returns:
        str: New version if available, None if not
    """
    regex = r'\([0-9]+\.[0-9]+\.[0-9]+\)'
    try:
        response = subprocess.check_output(
            'yolk -U dash-tools',
            shell=True,
            stderr=subprocess.DEVNULL)
        version = re.search(regex, response.decode('utf-8'))
        return version.group(0) if version else None
    except subprocess.CalledProcessError:
        # Problem with checking for updates; fail gracefully
        cprint(
            f"WARNING: There was an error checking the latest version of dash-tools.", 'yellow')
        return None


def check_for_updates():
    """
    Check for new version of dash-tools and prompt update if available
    """
    new_version = _new_version_available()
    if new_version:
        cprint(
            f"INFO: You are using dash-tools version {__version__}, however version {new_version.strip('(').strip(')')} is available.\nINFO: You should consider upgrading via the 'python -m pip install --upgrade dash-tools' command.", 'yellow')



'''
 # @ Author: Andrew Hossack
 # @ Create Time: 2022-05-23 12:48:14
 # Generate interesting app names
'''

import random
from pkg_resources import resource_filename
import os


def _get_data_path(filename: os.PathLike) -> os.PathLike:
    """
    Get the path to the data directory.
    """
    return resource_filename(__name__, filename)


def get_words(N: int) -> list:
    """
    Get N random words from the data file.
    """
    with open(_get_data_path('nounlist.csv'), 'r') as f:
        words = f.read().splitlines()
    return [words[i] for i in sorted(random.sample(range(len(words)), N))]



'''
 # @ Author: Andrew Hossack
 # @ Create Time: 2022-04-28 11:46:39
 # File util functions
'''
import os
import re
from typing import Union
from pipreqs import pipreqs


def check_file_exists(root_path: os.PathLike, file_name: str) -> bool:
    """
    Check that a file exists
    """
    path = os.path.join(root_path, file_name)
    if not os.path.exists(path):
        return False
    return True


def _add_requirement(root_path: os.PathLike, requirement: str):
    """
    Adds a requirement to the requirements.txt file if it doesn't already exist
    """
    # Append gunicorn to requirements.txt
    with open(os.path.join(root_path, 'requirements.txt'), 'r+') as requirements_file:
        if requirement not in requirements_file.read():
            requirements_file.write(f'{requirement}\n')


def create_requirements_txt(root_path: os.PathLike, destination: os.PathLike = None, update=False):
    """
    Creates requirements.txt file using pipreqs

    Args:
        root_path: Path to the root directory of the project
        destination: Optional path to save reqs file
        update: Optional boolean to update existing requirements.txt
    """
    print(
        f'dashtools: {"Found requirements.txt. Updating" if update else "Creating requirements.txt"} to {root_path}...')
    try:
        args = {'<path>': root_path, '--encoding': 'utf8', '--pypi-server': None, '--proxy': None, '--proxy-auth': None,
                '--proxy-type': None, '--requirement': None, '--requirements': None, '--version': None, '--use-local': None, '--savepath': None, '--diff': None, '--clean': None, '--print': None, '--force': None, '--mode': None}
        args['--savepath'] = os.path.join(destination,
                                          'requirements.txt') if destination else None
        if update:
            args['--force'] = True
            pipreqs.init(args)
        else:
            pipreqs.init(args)
    except Exception as e:
        # Catch any pipreqs exceiptions
        print(e)
        print('dashtools: Error creating requirements.txt')
        print('dashtools: Are you in a valid dash app directory?')
        exit('dashtools: Exiting')

    # Add requirements that might not be in requirements.txt
    for req in ['gunicorn', 'pandas', 'dash-tools']:
        _add_requirement(destination if destination else root_path, req)


def create_runtime_txt(root_path: os.PathLike):
    """
    Create runtime.txt file
    Default behavior is to use python-3.10.7
    """
    with open(os.path.join(root_path, 'runtime.txt'), 'w') as runtime_file:
        runtime_file.write('python-3.10.7')
    print('dashtools: Created runtime.txt using python-3.10.7')


def app_root_path(root_path: os.PathLike) -> Union[os.PathLike, None]:
    """
    Look for an app.py file in the directory recursively

    Returns:
        Path to app.py file if found, else None
    """
    # Find app.py file in the root_path directory
    app_path = None
    for root, _, files in os.walk(root_path):
        if 'app.py' in files:
            app_path = root
            break
    if app_path is None:
        print('dashtools: Error: No app.py file found! An app.py file is needed for this operation.')
        exit('dashtools: Exiting')
    return app_path


def create_procfile(root_path: os.PathLike):
    """
    Create a procfile.
    """
    app_path = app_root_path(root_path)
    rel_path = os.path.relpath(app_path, root_path)
    chdir = f'{" --chdir " + rel_path if rel_path != "." else ""}'
    with open(os.path.join(root_path, 'Procfile'), 'w') as procfile:
        procfile.write(
            f'web: gunicorn{chdir if len(chdir) > 0 else ""} app:server')
    print(f'dashtools: Created Procfile')


def verify_procfile(root_path: os.PathLike) -> dict:
    """
    Verifies that the Procfile is correct:

    Ex. If "... --chdir src app:server ..." is in Procfile, check that
        'server' hook exists in src/app.py

    Returns:
        {
            'valid': True (valid) or False (invalid),
            'dir': Directory of app,
            'module': Module name
            'hook': Hook name
        }
    """
    with open(os.path.join(root_path, 'Procfile'), 'r', encoding="utf8") as procfile:
        procfile_contents = procfile.read()

    # Look for --chdir somedir
    chdir_regex = r"--chdir [a-zA-Z\/\\]+"
    try:
        chdir = re.search(chdir_regex, procfile_contents).group(0)
        chdir = chdir.replace('--chdir ', '')
    except (AttributeError, IndexError):
        chdir = ''

    # Look for module:hook
    hook_regex = r"[a-zA-Z]+:[a-zA-Z]+"
    try:
        hook = re.search(hook_regex, procfile_contents).group(0)
        hook_module = hook.split(':')[0] + '.py'
        hook = hook.split(':')[1]
    except (AttributeError, IndexError):
        hook = ''
        return {
            'valid': False,
            'dir': chdir,
            'hook': hook,
            'module': hook_module
        }

    modpath = os.path.join(root_path, chdir, hook_module)

    # Check that the module exists
    if not os.path.exists(modpath):
        return {
            'valid': False,
            'dir': chdir,
            'hook': hook,
            'module': hook_module
        }

    # Check that the hook exists in the module
    with open(modpath, 'r', encoding="utf8") as modfile:
        modfile_contents = modfile.read()

    # Look for the hook "{hook} =" or "{hook}=" with spaces and newlines
    hook_regex = f"^([\n]+{hook}\s=|[\n]+{hook}=|{hook}=|{hook}\s=)"
    try:
        re.search(hook_regex, modfile_contents, re.MULTILINE).group(0)
        return {
            'valid': True,
            'dir': chdir,
            'hook': hook,
            'module': hook_module
        }
    except (AttributeError, IndexError):
        return {
            'valid': False,
            'dir': chdir,
            'hook': hook,
            'module': hook_module
        }



'''
 # @ Author: Andrew Hossack
 # @ Create Time: 2022-04-01 13:57:48
 # Build dash apps with dashtools
'''

import datetime
import os
import shutil

from dashtools.templating import buildAppUtils, pipUtils


def _format_file(name: os.PathLike, app_name: str, dest: os.PathLike):
    """
    Look for files ending in .py, .md, Procfile, and format them

    Change {appName} to the app name
    Change {createTime} to the current time

    Modifies the user's file in place
    """
    if('.py' in name or '.md' in name or 'Procfile' in name):
        with open(dest, 'r') as f:
            content = f.read()
            content = content.replace(r'{appName}', app_name)
            content = content.replace(
                r'{createTime}', str(datetime.datetime.now()))
            with open(dest, 'w') as f:
                f.write(content)


def _get_template_path(template_value: str) -> str:
    """
    Get path to template directory
    """
    template_dir = os.path.join('templates', template_value)
    return buildAppUtils.get_templates_data_path(template_dir)


def create_app(target_dir: os.PathLike, app_name: str, template: buildAppUtils.Template):
    '''
    Create a new app in the target directory.

    Looks for files in the /template directory
    '''
    # Check arguments
    buildAppUtils.check_create_app_args(target_dir, app_name)

    # Check for file write permissions in the base directory (command invoke directory)
    if not buildAppUtils.check_write_permission(target_dir):
        print(
            f'dashtools: init: No write permissions for {target_dir}')
        exit(f'dashtools: init: Failed')

    # Copy files from template directory
    template_base_path = _get_template_path(template.value)
    for path, _, files in os.walk(template_base_path):
        for name in files:
            # Skip non .template files
            if('.template' not in name):
                continue

            # Get the relative path to the file
            relative_path = os.path.relpath(
                path, template_base_path)
            src = os.path.join(path, name)

            # Get the destination path
            rel_path = relative_path if relative_path != '.' else ''
            dest = os.path.join(target_dir, app_name, rel_path, name)
            dest = dest.replace('.template', '')
            dest = dest.replace(r'{appName}', app_name)

            # Create the directory if it doesn't exist
            if not os.path.exists(os.path.dirname(dest)):
                os.makedirs(os.path.dirname(dest))

            # Copy the file
            shutil.copyfile(src, dest)

            # Format the file
            _format_file(name, app_name, dest)

    # Handle pip requirements
    pipUtils.handle_template_requirements(template.value)
    print(
        f'dashtools: init: Created new app {app_name} at {os.path.join(target_dir, app_name)} with {template.name} template')



'''
 # @ Author: Andrew Hossack
 # @ Create Time: 2022-06-19 12:25:30
'''


import os
import subprocess
import webbrowser

from dashtools.deploy import fileUtils
from dashtools.deploy.deployHeroku import prompt_user_choice


def _check_docker_installed() -> bool:
    """
    Check if docker is installed.

    :return: True if docker is installed, False otherwise.
    """
    try:
        subprocess.check_output(
            f'docker --help',
            shell=True,
            stderr=subprocess.DEVNULL)
        return True
    except subprocess.CalledProcessError:
        return False


def _write_dockerfile(root_dir: os.PathLike, destination_dir: os.PathLike) -> None:
    """
    Write a Dockerfile to the specified directory.

    :param root_dir: The root directory of the project.
    :param destination_dir: The directory to write the Dockerfile to.
    """
    # Get relative path of app.py file
    rel_path = os.path.relpath(fileUtils.app_root_path(root_dir), root_dir)
    # TODO test if app.py file changes locations
    contents = f"""FROM python:3.9-slim
                COPY requirements.txt ./requirements.txt
                RUN pip install -r requirements.txt
                COPY . ./
                CMD gunicorn -b 0.0.0.0:80 {rel_path}.app:server"""
    with open(os.path.join(destination_dir, 'Dockerfile'), 'w') as f:
        for line in contents.split('\n'):
            f.write(line.strip() + '\n')


def create_image(image_name: str, cwd: os.PathLike) -> None:
    """
    Create a new docker image for the current project.

    :param destination_parent_dir: The directory to create the image in.
    :param image_name: The name of the app.
    :param cwd: The current working directory.
    """
    # 1. Check if docker is installed
    if not _check_docker_installed():
        print('dashtools: docker: error: Docker is not installed!')
        if prompt_user_choice('dashtools: Open Docker website?'):
            webbrowser.open('https://docs.docker.com/get-docker/')
        exit(1)

    # 2. Check if requirements.txt exists in image directory
    if not fileUtils.check_file_exists(cwd, 'requirements.txt'):
        print('dashtools: docker: error: No requirements.txt file found!')
        if prompt_user_choice('dashtools: Generate requirements.txt file?'):
            fileUtils.create_requirements_txt(cwd)
        else:
            exit(
                'dashtools: docker: error: A requirements.txt file is required to create an image!')
    else:
        print('dashtools: docker: init: Requirements file found')
        fileUtils.create_requirements_txt(
            cwd, update=True)

    # 3. Check if Dockerfile exists in image directory
    if not fileUtils.check_file_exists(cwd, 'Dockerfile'):
        print('dashtools: docker: error: No Dockerfile found')
        if prompt_user_choice('dashtools: Create a Dockerfile?'):
            _write_dockerfile(cwd, cwd)
        else:
            exit('dashtools: docker: error: Dockerfile needed to create image')

    # 4. Build image
    print(f'dashtools: docker: init: Building image {image_name}')
    try:
        subprocess.run(
            f'docker build -t {image_name} .',
            shell=True,
            check=True)
    except subprocess.CalledProcessError as e:
        print(e)
        exit('dashtools: docker: error: Failed to build image')

    print(f'dashtools: docker: init: Image {image_name} created!')
    print(
        f'dashtools: Run `docker run -p 8080:80 {image_name}` to start the image on http://localhost:8080')
    print('dashtools: Run `docker container ls -a` to see all images')



'''
 # @ Author: Andrew Hossack
 # @ Create Time: 2022-05-28 13:56:14
'''

import configparser
from typing import List, Union


def _get_config() -> List[str]:
    """
    Get the config file
    """
    config = configparser.ConfigParser()
    config.read('config.ini')
    return config


def get_config_value(key: str, section: str = "DEFAULT") -> Union[str, None]:
    """
    Get the value of a key in a section or None if not found
    """
    config = _get_config()
    try:
        value = config[section][key]
        if len(value) > 0:
            return value
        return None
    except KeyError:
        return None


def set_config_value(key: str, value: str, section: str = "DEFAULT") -> None:
    """
    Set the value of a key in a section
    """
    config = _get_config()
    config.set(section, key, value)
    with open('config.ini', 'w') as configfile:
        config.write(configfile)



'''
 # @ Author: Andrew Hossack
 # @ Create Time: 2022-05-23 00:08:07
 # @ Handle running the app
'''


import os
import subprocess
from typing import Union
from dashtools.data import configUtils
from dashtools.deploy import fileUtils


def _is_correct_python_command(command: str) -> bool:
    """
    Try running the python command
    Return True if it works, else False
    """
    try:
        subprocess.check_output(
            f'{command} --version',
            shell=True,
            stderr=subprocess.DEVNULL)
        return True
    except subprocess.CalledProcessError:
        return False


def _try_all_commands() -> Union[str, None]:
    """
    Try all python commands
    Return the first one that works or None if none work
    """
    command = None
    for cmd in ['python', 'python3', 'python.exe', 'python3.exe']:
        if _is_correct_python_command(cmd):
            command = cmd
            break
    return command


def set_python_shell_cmd(command: str):
    """
    Set the python shell command
    """
    if _is_correct_python_command(command):
        configUtils.set_config_value('python_shell_cmd', command)
        print(
            f'dashtools: run: success: Python shell command set to {command}')
    else:
        print(
            f'dashtools: run: error: Command {command} is not valid for running Python')
    return


def _python_shell_cmd() -> str:
    """
    Get the python shell command
    """
    command = configUtils.get_config_value('python_shell_cmd')
    if not command:
        command = _try_all_commands()
        if command:
            configUtils.set_config_value('python_shell_cmd', command)
        else:
            print('dashtools: run: error: None of the following commands were found on your system: python, python.exe, python3, python3.exe')
            print('dashtools: run: Please set the python shell command with "dashtools run --set-python-shell-cmd <command>", or fall back to running your app from the app.py file')
            exit('dashtools: View Troubleshooting docs for more info')
    return command


def _run_from_app(root_path: os.PathLike):
    """
    Try looking for an app.py file in the directory recursively
    """
    # Find app.py file in the root_path director
    root = fileUtils.app_root_path(root_path)
    if not root:
        exit('dashtools: run: No app.py file found')
    try:
        print(
            f'dashtools: Running From {root + "/" if root else ""}app.py')
        os.chdir(root)
        # Has to run as os.system() to get the output
        # This means that if no app file is found, an error will be printed to screen
        # And cannot be handled in a cleaner way
        os.system(f'{_python_shell_cmd()} app.py')
        return
    except Exception:
        pass


def run_app(root_path: os.PathLike):
    '''
    Look for a Procfile to run the app, else recursive search for app.py file
    '''
    # Check Procfile exists
    if fileUtils.check_file_exists(root_path, 'Procfile'):
        proc = fileUtils.verify_procfile(root_path)
        if proc['valid']:
            print('dashtools: Running From Procfile')
            os.chdir(root_path)
            modpath = proc["dir"].replace("/", ".").replace("\\", ".")
            modname = proc["module"].replace(".py", "")
            os.system(f'{_python_shell_cmd()} -m {modpath}.{modname}')
        else:
            # Invalid Procfile, try running app.py
            _run_from_app(root_path)
    else:
        # No Procfile, try running app.py
        _run_from_app(root_path)
    return



from version import __version__
# -- Project information

project = 'DashTools'
author = 'Andrew Hossack'
copyright = '2022, ' + author

version = __version__
release = __version__

# -- General configuration

html_static_path = ["_static"]
html_theme_options = {
    "light_logo": "images/logo_bk_small.png",
    "dark_logo": "images/logo_w_small.png",
    "top_of_page_button": None,
    "announcement": "<em>Thank you for using DashTools!</em> If you like it, consider leaving the project a <a href='https://github.com/andrew-hossack/dash-tools' target='_blank'>⭐ on GitHub</a>.",
}

extensions = [
    'sphinx.ext.duration',
    'sphinx.ext.doctest',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'sphinx_copybutton',  # copy button on code snippets
    'myst_parser',  # md parser
]

source_suffix = ['.rst', '.md']

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'sphinx': ('https://www.sphinx-doc.org/en/master/', None),
}
intersphinx_disabled_domains = ['std']

templates_path = ['_templates']

# -- Options for HTML output

html_theme = 'furo'
html_title = f'DashTools Documentation - v{__version__}'

# -- Options for EPUB output
epub_show_urls = 'footnote'



'''
 # @ Author: Andrew Hossack
 # @ Create Time: 2022-04-04 13:23:14
'''
import os
import re
import subprocess
import webbrowser

from dashtools.deploy import fileUtils, gitUtils, herokuUtils


def _handle_heroku_files(root_path: os.PathLike) -> bool:
    """
    Check for Procfile, runtime.txt, requirements.txt
    """
    procfile_exists = fileUtils.check_file_exists(root_path, 'Procfile')
    runtime_exists = fileUtils.check_file_exists(root_path, 'runtime.txt')
    requirements_exists = fileUtils.check_file_exists(
        root_path, 'requirements.txt')

    if not procfile_exists:
        if prompt_user_choice('dashtools: Required file Procfile not found. Create one automatically?'):
            fileUtils.create_procfile(root_path)
            procfile_exists = True

    if not runtime_exists:
        if prompt_user_choice('dashtools: Required file runtime.txt not found. Create one automatically?'):
            fileUtils.create_runtime_txt(root_path)
            runtime_exists = True

    if not requirements_exists:
        if prompt_user_choice('dashtools: Required file requirements.txt not found. Create one automatically?'):
            fileUtils.create_requirements_txt(root_path, update=False)
            requirements_exists = True
    else:
        # Update requirements.txt
        fileUtils.create_requirements_txt(root_path, update=True)

    if not (procfile_exists and runtime_exists and requirements_exists):
        print('dashtools: Procfile, runtime.txt, and requirements.txt are needed for Heroku deployment.')
        exit('dashtools: heroku: deploy: Aborted')


def prompt_user_choice(message: str, prompt: str = 'Continue? (y/n) > ', does_repeat: bool = True) -> bool:
    """
    Prompt the user to continue or not.

    Args:
        message: Message to display to the user
        prompt: Prompt to display to the user
        does_repeat: If True, the user will be prompted again if they enter an invalid response

    Returns:
        True (y/Y)
        False (n/N)
    """
    print(message)
    response = input(f'dashtools: {prompt}')
    if response.lower() == 'y':
        return True
    elif response.lower() == 'n':
        return False
    elif does_repeat:
        return prompt_user_choice(message)
    else:
        return False


def _add_changes_and_push_to_heroku(heroku_app_name: str, remote: str = 'heroku') -> bool:
    """
    Add changes to the repository and push to Heroku

    Returns:
        True (Success)
        False (Failure)
    """
    # Create a commit to push to Heroku
    try:
        subprocess.check_output(f'git add .', shell=True)
        print(f'dashtools: Created commit to push to {remote}')
        subprocess.check_output(
            f'git commit -m "Deploy to Heroku for app {heroku_app_name} - dashtools"', shell=True)
        # Push to Heroku
        print(f'dashtools: Pushing to Heroku')
        subprocess.check_output(f'git push {remote} HEAD:master', shell=True)
    except subprocess.CalledProcessError:
        return False
    return True


def _remove_heroku_remote():
    """
    Remove the heroku remote

    Returns:
        True if successful
        False if failed
    """
    print('dashtools: Removing existing heroku remote')
    try:
        subprocess.check_output('git remote rm heroku', shell=True)
    except subprocess.CalledProcessError:
        exit('dashtools: heroku: deploy: Failed to remove heroku remote')


def _check_heroku_remote_already_exists() -> bool:
    """
    Check if the heroku remote is set

    Returns:
        True if already set
        False if remote "heroku" is not set
    """
    regex = r'heroku'
    try:
        output = subprocess.check_output('git remote', shell=True)
        heroku_grep = re.search(regex, output.decode('utf-8'))
        try:
            heroku_grep.group(0)
        except AttributeError:
            return False
    except subprocess.CalledProcessError:
        exit('dashtools: heroku: deploy: Failed to check if heroku remote is set')
    return True


def _success_message(heroku_app_name: str):
    """
    Print a success message
    """
    print(f'\n\ndashtools: Successfully deployed to Heroku from git branch "heroku"!')
    print(
        f'dashtools: To push changes, select the "Update Existing App" option after typing: dashtools --heroku: deploy')
    print(
        f'dashtools: Management Page: https://dashboard.heroku.com/apps/{heroku_app_name}')
    print(
        f'dashtools: Application Page: https://{heroku_app_name}.herokuapp.com/')

    # Prompt user to open the deployed app
    if input('dashtools: Enter any key to open in browser or q to exit > ') != 'q':
        webbrowser.open(f'https://{heroku_app_name}.herokuapp.com/')

    print('dashtools: heroku: deploy: Finished')


def update_heroku_app(project_root: os.PathLike, remote: str = 'heroku'):
    """
    Updates the existing heroku app

    Args:
        remote(str): Remote to update. Default 'heroku'
    """
    # (Re)generate requirements file
    if fileUtils.check_file_exists(project_root, 'requirements.txt'):
        fileUtils.create_requirements_txt(project_root, update=True)
    if not _add_changes_and_push_to_heroku('update', remote=remote):
        print(f'dashtools: Unable to update heroku app. Project must be deployed to heroku before updating.')
        exit('dashtools: heroku: update: Failed')
    print('dashtools: Changes pushed to {remote} remote')
    exit('dashtools: heroku: update: Success')


def _get_valid_app_name() -> str:
    """
    Returns a unique and valid heroku app name
    """
    should_loop = True
    while should_loop:
        # Generate new name
        heroku_app_name = herokuUtils.get_heroku_app_name()
        # Check if the project already exists on Heroku if name is specified
        if not herokuUtils.check_heroku_app_name_available(heroku_app_name):
            print(
                f'dashtools: App "{heroku_app_name}" already exists on Heroku!')
            print('dashtools: Please choose a unique name that isn\'t already taken.')
        # Validate app name
        elif not herokuUtils.validate_heroku_app_name(heroku_app_name):
            print(
                f'dashtools: App name "{heroku_app_name}" is not valid!')
            print('dashtools: Heroku app names must start with a letter, end with a letter or digit, can only contain lowercase letters, numbers, and dashes, and have a minimum length of 3 characters and maximum length of 30 characters.')
        else:
            should_loop = False
    return heroku_app_name


def deploy_app_to_heroku(project_root_dir: os.PathLike):
    """
    Uses the Heroku CLI to deploy the current project
    """
    print('dashtools: heroku: deploy: Starting')

    # Check if heroku CLI is installed
    if not herokuUtils.heroku_is_installed():
        print(f'dashtools: Heroku CLI not installed!')
        print(f'dashtools: See https://devcenter.heroku.com/articles/heroku-cli#install-the-heroku-cli')
        # Prompt user to open the deployed app
        if input('dashtools: Enter any key to open in browser or q to exit > ') != 'q':
            webbrowser.open(
                'https://devcenter.heroku.com/articles/heroku-cli#install-the-heroku-cli')
        exit('dashtools: heroku: deploy: Failed')

    # Check if git is installed
    if not gitUtils.git_is_installed():
        print(f'dashtools: Git not installed!')
        print(f'dashtools: See https://git-scm.com/downloads')
        # Prompt user to open the deployed app
        if input('dashtools: Enter any key to open in browser or q to exit > ') != 'q':
            webbrowser.open('https://git-scm.com/downloads')
        exit('dashtools: heroku: deploy: Failed')

    # Check that git is initialized in the current repo
    if not gitUtils.is_git_repository():
        print(f'dashtools: A git repository is needed for deployment. Current directory is not a git repository!')
        # Prompt user to init git
        if prompt_user_choice('dashtools: Would you like to init git?'):
            os.system('git init')
        else:
            print('dashtools: To start a git repository, type: git init')
            exit('dashtools: heroku: deploy: Failed')

    # Check that heroku remote is not already set
    if _check_heroku_remote_already_exists():
        print(f'dashtools: Git remote "heroku" is already set!')
        print('dashtools: Please choose an option below:')
        print('\t1. Push to the existing heroku remote (Update Existing App)')
        print('\t2. Remove the heroku remote and continue (Create New App)')
        print('\t3. Abort')
        while True:
            response = input('dashtools: Choice (1, 2, 3) > ')
            if response == '1':
                update_heroku_app(project_root_dir)
            elif response == '2':
                _remove_heroku_remote()
                break
            elif response == '3':
                exit('dashtools: heroku: deploy: Aborted')
            else:
                pass

    # Check app.py file exists
    if not fileUtils.app_root_path(project_root_dir):
        print(f'dashtools: No app.py file found! An app.py file is required to deploy your app.')
        exit('dashtools: heroku: deploy: Failed')

    # Check that the project has necessary files, and generate them if not
    _handle_heroku_files(project_root_dir)

    # Get a unique app name
    heroku_app_name = _get_valid_app_name()

    # Check procfile is correct
    procfile = fileUtils.verify_procfile(project_root_dir)
    if not procfile['valid']:
        print(f'dashtools: Invalid Procfile!')
        print(
            f'dashtools: Did you include "{procfile["hook"]} = app.server" after instantiating "app = Dash(...)" in {procfile["dir"]}/{procfile["module"]}?')
        if not prompt_user_choice('dashtools: App will not start without a valid Procfile. Continuing is not recommended.'):
            exit('dashtools: heroku: deploy: Aborted')

    # Log into Heroku
    if not herokuUtils.login_heroku_successful():
        print(f'dashtools: Heroku login failed.')
        exit('dashtools: heroku: deploy: Failed')

    # Confirm deployment settings and create the project on Heroku if the user confirms
    if not prompt_user_choice(f'dashtools: Please confirm creating app {heroku_app_name} on Heroku and adding git remote "heroku".'):
        exit('dashtools: heroku: deploy: Aborted')

    # Create the project on Heroku and capture the git remote URL
    if not herokuUtils.create_app_on_heroku(heroku_app_name):
        exit(
            f'dashtools: heroku: deploy: Deploying app {heroku_app_name} to Heroku failed.')

    # Add python buildpack
    print(f'dashtools: Adding python buildpack')
    os.system(f'heroku buildpacks:add heroku/python -a {heroku_app_name}')

    # Push to Heroku
    _add_changes_and_push_to_heroku(heroku_app_name)

    # Print success message and exit
    _success_message(heroku_app_name)



'''
 # @ Author: Andrew Hossack
 # @ Create Time: 2022-04-28 11:37:09
 # Heroku util functions
'''
import random
import re
import string
import subprocess
from dashtools.data import randomWords
import requests


def heroku_is_installed() -> bool:
    """
    Check that heroku CLI is installed on the system
    Looks for "heroku/X.Y.Z " in the output of "heroku --version"
    """
    regex = r'heroku/[0-9]+\.[0-9]+\.[0-9]+ [a-zA-Z]+'
    isInstalled = True
    try:
        heroku_command_output = subprocess.check_output(
            'heroku --version', shell=True)
        if not re.search(regex, heroku_command_output.decode('utf-8')):
            isInstalled = False
    except subprocess.CalledProcessError:
        isInstalled = False
    return isInstalled


def login_heroku_successful() -> bool:
    """
    Try to log into Heroku
    """
    print(f'dashtools: Logging into Heroku...')
    regex = r'Error: quit'
    loginWasSuccessful = True
    try:
        heroku_command_output = subprocess.check_output(
            'heroku login', shell=True)
        if re.search(regex, heroku_command_output.decode('utf-8')):
            loginWasSuccessful = False
    except subprocess.CalledProcessError:
        loginWasSuccessful = False
    return loginWasSuccessful


def check_heroku_app_name_available(heroku_app_name: str) -> bool:
    """
    Check that the app name is available on Heroku

    Returns:
        True if available
        False if not
    """
    response = requests.get(
            url=f'https://{heroku_app_name}.herokuapp.com/',
            headers={'User-Agent': 'Custom'}
        )
    if response.status_code == 404:
        return True
    else:
        return False

def create_app_on_heroku(app_name: str) -> bool:
    """
    Create the project on Heroku and return the git remote URL

    Returns:
        Success of the operation (True/False)
    """
    print(f'dashtools: Creating {app_name} on Heroku')
    url_regex = r'https:\/\/git\.heroku\.com\/[a-zA-Z0-9-_]*\.git'
    try:
        heroku_command_output = subprocess.check_output(
            f'heroku create {app_name}', shell=True)
        git_remote_url = re.search(
            url_regex, heroku_command_output.decode('utf-8'))
        try:
            git_remote_url.group(0)
        except AttributeError:
            # No url found. Something is wrong
            return False
    except subprocess.CalledProcessError:
        return False
    return True


def _generate_app_name():
    # Generate a random app name with three words
    app_name = '-'.join(randomWords.get_words(3))
    # Append alphanumeric characters to the end of the app name
    # eg. apple-banana-pear-1g3f
    app_name += f"-{''.join(random.choices(string.ascii_lowercase + string.digits, k=4))}"
    app_name = app_name.lower()
    return app_name


def get_heroku_app_name():
    """
    Create or generate app name if one isn't provided
    """
    print('dashtools: Please type a unique app name or press enter to generate one automatically.')
    app_name = input('dashtools: App Name (Optional) > ')
    if app_name == '':
        while True:
            app_name = _generate_app_name()
            if validate_heroku_app_name(app_name):
                print(f'dashtools: Generated app name "{app_name}"')
                if input('dashtools: Continue with this name or regenerate?\ndashtools: (Continue: y, Regenerate: n) > ') == 'y':
                    break
    else:
        print(f'dashtools: Using app name "{app_name}"')
    return app_name


def validate_heroku_app_name(heroku_app_name) -> bool:
    """
    Heroku app names must start with a letter, end with a
    letter or digit, can only contain lowercase letters,
    numbers, and dashes, and have a minimum length of 3 characters.
    Maximum 30 characters.

    Returns:
        True if valid
        False if invalid
    """
    regex = r'^[a-z][a-z0-9-]{2,29}$'
    if re.search(regex, heroku_app_name):
        if heroku_app_name[-1] == '-':
            return False
        return True
    return False



'''
 # @ Author: Andrew Hossack
 # @ Create Time: 2022-05-24 17:46:41
'''

import subprocess
from dashtools.templating import buildApp
from dashtools.deploy import deployHeroku
import os
import pkg_resources


def _check_pip_installed() -> bool:
    """
    Check if pip is installed
    """
    try:
        subprocess.check_output(
            'pip --version',
            shell=True,
            stderr=subprocess.DEVNULL)
        return True
    except subprocess.CalledProcessError:
        return False


def _install_pip_requirement(requirement: str):
    '''
    Install a pip requirement
    '''
    print(f'dashtools: Installing {requirement}')
    os.system(f'pip install {requirement}')


def _check_pip_requirement_installed(requirement: str) -> bool:
    '''
    Check if a pip requirement is installed
    '''
    try:
        pkg_resources.get_distribution(requirement)
    except pkg_resources.DistributionNotFound:
        return False
    except pkg_resources.VersionConflict:
        return False
    return True


def _check_pip_requirements(requirements: list, template_value: str):
    '''
    Check if pip requirements are installed
    '''
    for req in requirements:
        if not _check_pip_requirement_installed(req):
            print(
                f'dashtools: Template {template_value} requires pip module {req}, which is not installed')
            if _check_pip_installed():
                if deployHeroku.prompt_user_choice(f'dashtools: Install {req}?'):
                    _install_pip_requirement(req)
                    print()
            else:
                print(f'dashtools: Please install {req} manually')


def _get_template_required_packages(template_value: str) -> list:
    """
    Looks for 'packages' file in template directory. If one is found,
    return a list of pip requirements.
    """
    packages_file = os.path.join(
        buildApp._get_template_path(template_value), 'packages.txt')
    if os.path.exists(packages_file):
        with open(packages_file, 'r') as f:
            return f.read().splitlines()
    return []


def handle_template_requirements(template_value: str):
    """
    Handle pip requirements for template
    """
    _check_pip_requirements(_get_template_required_packages(
        template_value), template_value)



'''
 # @ Author: Andrew Hossack
 # @ Create Time: 2022-04-04 13:41:57
 # @ Templates file
'''
from enum import Enum


class Template(Enum):
    """
    Enum of templates to be used in the app.
    Values must match the directory name in the templating/templates/ directory

    e.g.
    -> DEFAULT = 'default'
    -> Here the template 'DEFAULT' has a template directory named '/default':
    -> templates/default/
    """
    ADVANCED = 'advanced'
    CSV = 'csv'
    DEFAULT = 'default'
    FASTDASH = 'fastdash'
    LEAFLET = 'leaflet'
    IRIS = 'iris'
    MANTINE = 'mantine'
    MULTIPAGE = 'multipage'
    SIDEBAR = 'sidebar'
    TABS = 'tabs'



'''
 # @ Author: Andrew Hossack
 # @ Create Time: 2022-04-01 19:12:04
 '''
import argparse
from typing import Union
from pkg_resources import resource_filename
import os
from dashtools.templating.Templates import Template


def check_write_permission(path: os.PathLike) -> bool:
    """
    Check for write permission in directory/

    Returns:
        True if write access else allowed False
    """
    return os.access(path, os.W_OK)


def print_templates():
    """
    Print the available templates.
    """
    for template in Template:
        print(f'\t> {template.value}')


def check_create_app_args(base_dir, app_name):
    """
    Check the arguments passed to the init command.
    """
    app_dir = os.path.join(base_dir, app_name)
    if os.path.exists(app_dir):
        print(f'dashtools: init: App {app_dir} already exists.')
        print(
            f'dashtools: init: Please change the "{app_dir}" name or delete the {app_dir} directory!')
        exit(f'dashtools: init: Failed')


def get_templates_data_path(data_dir: os.PathLike) -> os.PathLike:
    """
    Get the path to the data directory.
    """
    return resource_filename(__name__, data_dir)


def get_template_from_args(args: argparse.Namespace) -> Union[str, Template]:
    """
    Get the template to use based on the arguments passed to the CLI.
    If the user passed in a template, use that, otherwise use the `default`.
    """
    possible_template = args.init[1] if len(
        args.init) > 1 else Template.DEFAULT
    return _convert_to_template_or_error(possible_template)


def _convert_to_template_or_error(value) -> Template:
    """
    Convert the string passed to the init command to a template.
    """
    try:
        return Template(value)
    except ValueError:
        print(f'dashtools: init: Template "{value}" is not a valid template.')
        print('Valid templates are:')
        print_templates()
        exit(1)



# -*- coding: utf-8 -*-
import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

meta = {}

with open('./dashtools/version.py') as f:
    exec(f.read(), meta)

with open('requirements.txt') as f:
    requirements = f.readlines()

setuptools.setup(
    name="dash-tools",
    version=meta['__version__'],
    author="Andrew Hossack",
    author_email="andrew_hossack@outlook.com",
    description="Plotly Dash Templating and Deployment Tools",
    entry_points={
        'console_scripts': ['dashtools = dashtools.cli.cli:main']},
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/andrew-hossack/dash-tools",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(),
    python_requires=">=3.6",
    include_package_data=True,
    install_requires=requirements,
)
